%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*
Nan.Validation 1.2
Validation Operations
nan_validation.pl

This file is part of
Nan.Common 1.2
Common Operations
<http://julio.diegidio.name/Projects/Prolog/Download/Nan.Common-0.1.zip>

Copyright 2012 J.P. Di Egidio
All rights reserved.
<http://julio.diegidio.name/Projects/Prolog/Nan.Common>

Nan.Common is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

Nan.Common is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with Nan.Common. If not, see <http://www.gnu.org/licenses/>.

As a special exception, if you link this library with other files,
compiled with a Free Software compiler, to produce an executable, this
library does not by itself cause the resulting executable to be covered
by the GNU General Public License. This exception does not however
invalidate any other reasons why the executable file might be covered by
the GNU General Public License.

Authors:
J.P. Di Egidio - JDE - mailto:julio@diegidio.name

History:
1.2.2012-11-11/JDE - Common: minor fixes to the docs and interface.
1.0.2012-11-10/JDE - Validation: created version 1.0.

Written and tested in:
SWI-Prolog 6.2.0 - <http://www.swi-prolog.org>
*/
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/**	<module> Nan.Validation 1.2: Validation Operations

*|Nan.Validation 1.2|*: Validation Operations (nan_validation)

This module is part of *|Nan.Common 1.2|*: Common Operations (nan_common)
<http://julio.diegidio.name/Projects/Prolog/Nan.Common>

This module provides common operations for testing and validation of term
types, domains and custom conditions, and for the construction of exception
and error terms.

This module abstracts and extends the type testing and validation system
provided by library(error).  In particular, this module extends to the types
that are documented in test_type/2.

NOTE: The type testing and validation system is extensible: specifically,
error:has_type/2 is a =multifile= (and =dynamic=) predicate.  See
library(error) for details.

@author		J.P. Di Egidio
@version	1.2
@copyright	Copyright 2012 J.P. Di Egidio
@license	GPL3 + SWI-Exception
@tbd		Extend the type system beyond error:has_type/2 (for type spec).
*/

:- module(nan_validation,
[	%%% INTERFACE: test_*
	test_type/2,					% +Type, @Term
	test_domain/3,					% +Type, :Constraint, @Term
	%%% INTERFACE: validate_*
	validate_type/3,				% +Type, @Term, +Message
	validate_type/4,				% +Type, @Term, :Caller, +Message
	validate_domain/4,				% +Type, :Constraint, @Term, +Message
	validate_domain/5,				% +Type, :Constraint, @Term, :Caller, +Message
	validate_condition/3,			% :Condition, +Error, +Message
	validate_condition/4,			% :Condition, +Error, :Caller, +Message
	%%% INTERFACE: *_exception
	type_exception/4,				% +Type, @Term, +Message, -Exception
	type_exception/5,				% +Type, @Term, :Caller, +Message, -Exception
	domain_exception/5,				% +Type, :Constraint, @Term, +Message, -Exception
	domain_exception/6,				% +Type, :Constraint, @Term, :Caller, +Message, -Exception
	custom_exception/3,				% +Error, +Message, -Exception
	custom_exception/4,				% +Error, :Caller, +Message, -Exception
	%%% INTERFACE: *_error
	type_error/3,					% +Type, @Term, -Error
	domain_error/4,					% +Type, :Constraint, @Term, -Error
	%%% INTERFACE: current_caller
	current_caller/1				% -Caller
]).

:- use_module(library(error), []).

:- meta_predicate
	%%% INTERFACE: test_*
	test_domain(+, 1, +),
	%%% INTERFACE: validate_*
	validate_type(+, +, :, +),
	validate_domain(+, 1, +, +),
	validate_domain(+, 1, +, :, +),
	validate_condition(0, +, +),
	validate_condition(0, +, :, +),
	%%% INTERFACE: *_exception
	type_exception(+, +, :, +, -),
	domain_exception(+, 1, +, +, -),
	domain_exception(+, 1, +, :, +, -),
	custom_exception(+, :, +, -),
	%%% INTERFACE: *_error
	domain_error(+, 1, +, -).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% META-IMPLEMENTATION:

expand_term_(Head, TermIn, TermOut) :-
	callable(TermIn),
	!,
	expand_callable_(Head, TermIn, TermOut).
expand_term_(_, Term, Term).

expand_callable_(Head, GoalIn, GoalOut) :-		%%	throw_no_direct_call_/0
	functor(GoalIn, throw_no_direct_call_, 0),
	!,
	expand_get_caller_(Head, Caller),
	GoalOut = (
		Error = context_error(no_direct_call, Head),
		Message = 'For term rewriting only.',
		Exception = error(Error, context(Caller, Message)),
		throw(Exception)
	).
expand_callable_(Head, GoalIn, GoalOut) :-		%%	validate_type/3
	functor(GoalIn, validate_type, 3),
	!,
	expand_get_caller_(Head, Caller),
	GoalIn = validate_type(Type, Term, Message),
	GoalOut = validate_type(Type, Term, Caller, Message).
%
expand_callable_(Head, GoalIn, GoalOut) :-		%%	validate_domain/4
	functor(GoalIn, validate_domain, 4),
	!,
	expand_get_caller_(Head, Caller),
	GoalIn = validate_domain(Type, Constraint, Term, Message),
	GoalOut = validate_domain(Type, Constraint, Term, Caller, Message).
%
expand_callable_(Head, GoalIn, GoalOut) :-		%%	validate_condition/3
	functor(GoalIn, validate_condition, 3),
	!,
	expand_get_caller_(Head, Caller),
	GoalIn = validate_condition(Condition, Error, Message),
	GoalOut = validate_condition(Condition, Error, Caller, Message).
%
expand_callable_(Head, GoalIn, GoalOut) :-		%%	type_exception/4
	functor(GoalIn, type_exception, 4),
	!,
	expand_get_caller_(Head, Caller),
	GoalIn = type_exception(Type, Term, Message, Exception),
	GoalOut = type_exception(Type, Term, Caller, Message, Exception).
%
expand_callable_(Head, GoalIn, GoalOut) :-		%%	domain_exception/5
	functor(GoalIn, domain_exception, 5),
	!,
	expand_get_caller_(Head, Caller),
	GoalIn = domain_exception(Type, Constraint, Term, Message, Exception),
	GoalOut = domain_exception(
		Type, Constraint, Term, Caller, Message, Exception
	).
%
expand_callable_(Head, GoalIn, GoalOut) :-		%%	custom_exception/3
	functor(GoalIn, custom_exception, 3),
	!,
	expand_get_caller_(Head, Caller),
	GoalIn = custom_exception(Error, Message, Exception),
	GoalOut = custom_exception(Error, Caller, Message, Exception).
%
expand_callable_(Head, GoalIn, GoalOut) :-		%%	current_caller/1
	functor(GoalIn, current_caller, 1),
	GoalIn = current_caller(CallerVar),
	var(CallerVar),
	!,
	expand_get_caller_(Head, Caller),
	GoalOut = (CallerVar = Caller).
%
expand_callable_(Head, GoalIn, GoalOut) :-
	GoalIn =.. [GName| GArgsIn],
	expand_callable__loop(Head, GArgsIn, GArgsOut),
	GoalOut =.. [GName| GArgsOut].

expand_callable__loop(Head, [GAIn| GArgsIn], [GAOut| GArgsOut]) :-
	!,
	expand_term_(Head, GAIn, GAOut),
	expand_callable__loop(Head, GArgsIn, GArgsOut).
expand_callable__loop(_, [], []).

expand_get_caller_(Head, Caller) :-
	context_module(Module),
	functor(Head, Name, Arity),
	Caller = Module:Name/Arity.

%%%	system:term_expansion(+TermIn, -TermOut) is det.

:- multifile
	system:term_expansion/2.

system:term_expansion((Head :- BodyIn), (Head :- BodyOut)) :-
	nan_validation:expand_term_(Head, BodyIn, BodyOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% INTERFACE: test_*

%%	test_type(
%%		+Type:type,
%%		@Term:any
%%	) is semidet.
%
%	Succeeds if Term satisfies Type, otherwise fails.
%
%	Supports all types documented in error:must_be/2 plus those listed here:
%
%	| =predicate=						| =atom=/=nonneg=													|
%	| =formatted=						| =text=-=list= _or_ =text=											|
%	| =formal_error=					| =callable=														|
%	| =term_order=						| =oneof=([=, <, >])												|
%	| =pair=(=KeyType=-=ValueType=)		| =KeyType=-=ValueType=												|
%	| =or=([=Type1=, ..., =TypeN=])		| =Type1= _or_ ... _or_ =TypeN=										|
%	| =var=(=SubType=)					| =var= _or_ =SubType=												|
%	| =m_var=(=SubType=)				| =atom=:=var= _or_ =atom=:=SubType=								|
%	| =m_predicate=						| =atom=:=predicate= _or_ =predicate=								|
%	| =m_callable=						| =atom=:=callable= _or_ =callable=									|
%	| =s_callable=(_Struct_:=struct=)	| =callable= _and_ =has_struct=(_Struct_, Term) (*)					|
%	| =struct=							| =callable= _with_ "_terminals_" =atom= _or_ =var=:=SubType= (*)	|
%	| =type=							| =callable= _and_ =clause=(=error=:=has_type=(Term, _))			|
%
%	(*) The implicit predicate has_struct/2 tests functor equality of
%	non-"terminal" sub-terms of _Struct_ to the corresponding sub-terms of
%	Term. For "terminal" sub-terms of _Struct_, specifically meaning any
%	encountered (sub-)terms of _Struct_ that are of type =var=:=SubType=,
%	has_struct/2 tests that the corresponding (sub-)term of Term satisfies
%	=SubType=.  (Variable names in "terminals" of _Struct_ are ignored but
%	useful for self-documentation.)
%	
%	@param	Type			The type that Term must satisfy.
%	@param	Term			The term to test.
%	@error	type_error: invalid argument(s).
%	@tbd	Extend type =formal_error= (to ISO?).
%	@tbd	Extend type =type= to a proper type specification
%			(e.g. =|type_spec(?Type, ?Struct, ?Test, ?Validate)|=) (?).

test_type(Type, Term) :-
	validate_arg_type_(Type),
	test_type_(Type, Term).

%%	test_domain(
%%		+Type:type,
%%		:Constraint:m_callable,
%%		@Term:any
%%	) is semidet.
%
%	Succeeds if Term satisfies Type and Constraint, otherwise fails.
%
%	Supports the types documented in test_type/2.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Constraint		The constraint that Term must satisfy.
%	@param	Term			The term to test.
%	@error	type_error: invalid argument(s).

test_domain(Type, Constraint, Term) :-
	validate_arg_type_(Type),
	validate_arg_constraint_(Constraint),
	test_domain_(Type, Constraint, Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% INTERFACE: validate_*

%%	validate_type(
%%		+Type:type,
%%		@Term:any,
%%		+Message:var(formatted)
%%	) is det.
%%
%%	validate_type(
%%		+Type:type,
%%		@Term:any,
%%		:Caller:m_var(predicate),
%%		+Message:var(formatted)
%%	) is det.
%
%	Succeeds if Term satisfies Type, otherwise throws
%	=|error(type_error(Type, Term), context(Caller, Message))|=.
%
%	Supports the types documented in test_type/2.
%
%	NOTE: The variant validate_type/3 is translated to
%	validate_type/4 via term rewriting,
%	with Caller set to the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Term			The term to test.
%	@param	Caller			(Optional) A predicate indicator representing the caller.
%	@param	Message			(Optional) An error message.
%	@error	type_error: invalid argument(s).
%	@error	type_error(Type, Term)

validate_type(_, _, _) :-
	throw_no_direct_call_.

validate_type(Type, Term, Caller, Message) :-
	validate_arg_type_(Type),
	validate_arg_caller_(Caller),
	validate_arg_message_(Message),
	validate_type_(Type, Term, Caller, Message).

%%	validate_domain(
%%		+Type:type,
%%		:Constraint:m_callable,
%%		@Term:any,
%%		+Message:var(formatted)
%%	) is det.
%%
%%	validate_domain(
%%		+Type:type,
%%		:Constraint:m_callable,
%%		@Term:any,
%%		:Caller:m_var(predicate),
%%		+Message:var(formatted)
%%	) is det.
%
%	Succeeds if Term satisfies Type and Constraint, otherwise throws
%	=|error(Error, context(Caller, Message))|=,
%	where _Error_ is =|type_error(Type, Term)|= or
%	=|domain_error(constr(Type, Constraint), Term)|=,
%	depending on the case.
%
%	Supports the types documented in test_type/2.
%
%	NOTE: The variant validate_domain/4 is translated to
%	validate_domain/5 via term rewriting,
%	with Caller set to the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Constraint		The constraint that Term must satisfy.
%	@param	Term			The term to test.
%	@param	Caller			(Optional) A predicate indicator representing the caller.
%	@param	Message			(Optional) An error message.
%	@error	type_error: invalid argument(s).
%	@error	type_error(Type, Term)
%	@error	domain_error(constr(Type, Constraint), Term)

validate_domain(_, _, _, _) :-
	throw_no_direct_call_.

validate_domain(Type, Constraint, Term, Caller, Message) :-
	validate_arg_type_(Type),
	validate_arg_constraint_(Constraint),
	validate_arg_caller_(Caller),
	validate_arg_message_(Message),
	validate_domain_(Type, Constraint, Term, Caller, Message).

%%	validate_condition(
%%		:Condition:m_callable,
%%		+Error:formal_error,
%%		+Message:var(formatted)
%%	) is det.
%%
%%	validate_condition(
%%		:Condition:m_callable,
%%		+Error:formal_error,
%%		:Caller:m_var(predicate),
%%		+Message:var(formatted)
%%	) is det.
%
%	Succeeds if calling Condition succeeds, otherwise throws
%	=|error(Error, context(Caller, Message))|=.
%
%	NOTE: The variant validate_condition/3 is translated to
%	validate_condition/4 via term rewriting,
%	with Caller set to the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	@param	Condition		The goal representing the validation condition.
%	@param	Error			The formal error to throw if Condition fails.
%	@param	Caller			(Optional) A predicate indicator representing the caller.
%	@param	Message			(Optional) An error message.
%	@error	type_error: invalid argument(s).
%	@error	Error

validate_condition(_, _, _) :-
	throw_no_direct_call_.

validate_condition(Condition, Error, Caller, Message) :-
	validate_arg_condition_(Condition),
	validate_arg_error_(Error),
	validate_arg_caller_(Caller),
	validate_arg_message_(Message),
	validate_condition_(Condition, Error, Caller, Message).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% INTERFACE: *_exception

%%	type_exception(
%%		+Type:type,
%%		@Term:any,
%%		+Message:var(formatted),
%%		-Exception:exception
%%	) is det.
%%
%%	type_exception(
%%		+Type:type,
%%		@Term:any,
%%		:Caller:m_var(predicate),
%%		+Message:var(formatted),
%%		-Exception:exception
%%	) is det.
%
%	Constructs an exception of the form
%	=|error(type_error(Type, Term), context(Caller, Message))|=.
%
%	Supports the types documented in test_type/2.
%
%	NOTE: The variant type_exception/4 is translated to
%	type_exception/5 via term rewriting,
%	with Caller set to the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Term			The term to test.
%	@param	Caller			(Optional) A predicate indicator representing the caller.
%	@param	Message			(Optional) An error message.
%	@param	Exception		The resulting exception term.
%	@error	type_error: invalid argument(s).

type_exception(_, _, _, _) :-
	throw_no_direct_call_.

type_exception(Type, Term, Caller, Message, Exception) :-
	validate_arg_type_(Type),
	validate_arg_caller_(Caller),
	validate_arg_message_(Message),
	type_exception_(Type, Term, Caller, Message, Exception).

%%	domain_exception(
%%		+Type:type,
%%		:Constraint:m_callable,
%%		@Term:any,
%%		+Message:var(formatted),
%%		-Exception:exception
%%	) is det.
%%
%%	domain_exception(
%%		+Type:type,
%%		:Constraint:m_callable,
%%		@Term:any,
%%		:Caller:m_var(predicate),
%%		+Message:var(formatted),
%%		-Exception:exception
%%	) is det.
%
%	Constructs an exception of the form
%	=|error(Error, context(Caller, Message))|=,
%	where _Error_ is
%	=|domain_error(constr(Type, Constraint), Term)|=.
%
%	Supports the types documented in test_type/2.
%
%	NOTE: The variant domain_exception/5 is translated to
%	domain_exception/6 via term rewriting,
%	with Caller set to the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Constraint		The constraint that Term must satisfy.
%	@param	Term			The term to test.
%	@param	Caller			(Optional) A predicate indicator representing the caller.
%	@param	Message			(Optional) An error message.
%	@param	Exception		The resulting exception term.
%	@error	type_error: invalid argument(s).

domain_exception(_, _, _, _, _) :-
	throw_no_direct_call_.

domain_exception(Type, Constraint, Term, Caller, Message, Exception) :-
	validate_arg_type_(Type),
	validate_arg_constraint_(Constraint),
	validate_arg_caller_(Caller),
	validate_arg_message_(Message),
	domain_exception_(Type, Constraint, Term, Caller, Message, Exception).

%%	custom_exception(
%%		+Error:formal_error,
%%		+Message:var(formatted),
%%		-Exception:exception
%%	) is det.
%%
%%	custom_exception(
%%		+Error:formal_error,
%%		:Caller:m_var(predicate),
%%		+Message:var(formatted),
%%		-Exception:exception
%%	) is det.
%
%	Constructs an exception of the form
%	=|error(Error, context(Caller, Message))|=.
%
%	NOTE: The variant custom_exception/3 is translated to
%	custom_exception/4 via term rewriting,
%	with Caller set to the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	@param	Error			The formal error.
%	@param	Caller			(Optional) A predicate indicator representing the caller.
%	@param	Message			(Optional) An error message.
%	@param	Exception		The resulting exception term.
%	@error	type_error: invalid argument(s).

custom_exception(_, _, _) :-
	throw_no_direct_call_.

custom_exception(Error, Caller, Message, Exception) :-
	validate_arg_error_(Error),
	validate_arg_caller_(Caller),
	validate_arg_message_(Message),
	custom_exception_(Error, Caller, Message, Exception).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% INTERFACE: *_error

%%	type_error(
%%		+Type:type,
%%		@Term:any,
%%		-Error:formal_error
%%	) is det.
%
%	Constructs a formal error of the form
%	=|type_error(Type, Term)|=.
%
%	Supports the types documented in test_type/2.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Term			The term to test.
%	@param	Error			The resulting error term.
%	@error	type_error: invalid argument(s).

type_error(Type, Term, Error) :-
	validate_arg_type_(Type),
	type_error_(Type, Term, Error).

%%	domain_error(
%%		+Type:type,
%%		:Constraint:m_callable,
%%		@Term:any,
%%		-Error:formal_error
%%	) is det.
%
%	Constructs a formal error of the form
%	=|domain_error(constr(Type, Constraint), Term)|=.
%
%	Supports the types documented in test_type/2.
%
%	@param	Type			The type that Term must satisfy.
%	@param	Constraint		The constraint that Term must satisfy.
%	@param	Term			The term to test.
%	@param	Error			The resulting error term.
%	@error	type_error: invalid argument(s).

domain_error(Type, Constraint, Term, Error) :-
	validate_arg_type_(Type),
	validate_arg_constraint_(Constraint),
	domain_error_(Type, Constraint, Term, Error).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% INTERFACE: current_caller

%%	current_caller(
%%		-Caller:m_callable
%%	) is det.
%
%	Returns in Caller the (qualified) predicate indicator of
%	the predicate containing the call.
%
%	NOTE: An invocation to this predicate is translated
%	via term rewriting as shown in the following example:
%	
%		For a predicate =my_mod:my_pred/1= such as:
%		
%		==
%		my_pred(X) :- ..., current_caller(Caller), ...
%		==
%		
%		The translation would be:
%		
%		==
%		my_pred(X) :- ..., Caller = my_mod:my_pred/1, ...
%		==
%
%	@param	Caller			The qualified predicate indicator for the caller.

current_caller(_) :-
	throw_no_direct_call_.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% IMPLEMENTATION:

%%%	error:has_type(+Type, @Term) is semidet.

:- multifile
	error:has_type/2.

error:has_type(predicate, Term) :-					% predicate
	nonvar(Term),
	Term = Name/Arity,
	atom(Name),
	error:has_type(nonneg, Arity).
error:has_type(formatted, Term) :-					% formatted
	nonvar(Term),
	(	Term = Format-Args
	->	error:has_type(list, Args)
	;	Term = Format
	),
	error:has_type(text, Format).
error:has_type(formal_error, Term) :-				% formal_error
	callable(Term).
error:has_type(term_order, Term) :-					% term_order
	error:has_type(oneof([=, <, >]), Term).
error:has_type(pair(KeyType-ValueType), Term) :-	% pair(KeyType-ValueType)
	nonvar(Term),
	Term = Key-Value,
	error:has_type(KeyType, Key),
	error:has_type(ValueType, Value).
error:has_type(or(Types), Term) :-					% or([Type1, ..., TypeN])
	error:has_type(list, Types),
	is_type_or_(Types, Term).
error:has_type(var(SubType), Term) :-				% var(SubType)
	(	var(Term)
	->	true
	;	error:has_type(SubType, Term)
	).
error:has_type(m_var(SubType), Term) :-				% m_var(SubType)
	stripped_module_(Term, SubTerm),
	(	var(SubTerm)
	->	true
	;	error:has_type(SubType, SubTerm)
	).
error:has_type(m_predicate, Term) :-				% m_predicate
	stripped_module_(Term, PredInd),
	error:has_type(predicate, PredInd).
error:has_type(m_callable, Term) :-					% m_callable
	stripped_module_(Term, Callable),
	callable(Callable).
error:has_type(s_callable(Struct), Term) :-			% s_callable(Struct)
	has_struct_(Struct, Term).
error:has_type(struct, Term) :-						% struct
	is_struct_(Term).
error:has_type(type, Term) :-						% type
	callable(Term),
	\+ \+ once(clause(error:has_type(Term, _), _)).

is_type_or_([Type| Types], Term) :-
	(	error:has_type(Type, Term)
	->	true
	;	is_type_or_(Types, Term)
	).

stripped_module_(Term, SubTerm) :-
	nonvar(Term),
	(	Term = Module:SubTerm
	->	atom(Module)
	;	Term = SubTerm
	).

is_struct_(Term) :-
	callable(Term),
	is_struct__case(Term).

is_struct__case(Term) :-
	atom(Term),
	!.
is_struct__case(Term) :-
	Term = Var:VType,
	var(Var),
	!,
	error:has_type(type, VType).
is_struct__case(Term) :-
	Term =.. [_| SubTerms],
	is_struct__loop(SubTerms).

is_struct__loop([Term| Terms]) :-
	is_struct_(Term),
	is_struct__loop(Terms).
is_struct__loop([]).

has_struct_(Struct, Term) :-
	callable(Struct),
	has_struct__case(Struct, Term).

has_struct__case(Struct, Term) :-
	atom(Struct),
	!,
	Term == Struct.
has_struct__case(Struct, Term) :-
	Struct = Var:VType,
	var(Var),
	!,
	error:has_type(VType, Term).
has_struct__case(Struct, Term) :-
	compound(Term),
	functor(Struct, Name, Arity),
	functor(Term, Name, Arity),
	Struct =.. [Name| SubStructs],
	Term =.. [Name| SubTerms],
	has_struct__loop(SubStructs, SubTerms).

has_struct__loop([Struct| Structs], [Term| Terms]) :-
	has_struct_(Struct, Term),
	has_struct__loop(Structs, Terms).
has_struct__loop([], []).

%%	validate_arg_type_(@Type) is det.
%%	validate_arg_constraint_(@Constraint) is det.
%%	validate_arg_condition_(@Condition) is det.
%%	validate_arg_error_(@Error) is det.
%%	validate_arg_caller_(@Caller) is det.
%%	validate_arg_message_(@Message) is det.

validate_arg_type_(Type) :-
	validate_type_(type, Type, _, _).

validate_arg_constraint_(Constraint) :-
	validate_type_(m_callable, Constraint, _, _).

validate_arg_condition_(Condition) :-
	validate_type_(m_callable, Condition, _, _).

validate_arg_error_(Error) :-
	validate_type_(formal_error, Error, _, _).

validate_arg_caller_(Caller) :-
	validate_type_(m_var(predicate), Caller, _, _).

validate_arg_message_(Message) :-
	validate_type_(var(formatted), Message, _, _).

%%	test_type_(+Type, @Term) is semidet.
%%	test_domain_(+Type, :Constraint, @Term) is semidet.

test_type_(Type, Term) :-
	error:has_type(Type, Term).

test_domain_(Type, Constraint, Term) :-
	error:has_type(Type, Term),
	call(Constraint, Term).

%%	validate_type_(+Type, @Term, :Caller, +Message) is det.
%%	validate_domain_(+Type, :Constraint, @Term, :Caller, +Message) is det.
%%	validate_condition_(:Condition, +Error, :Caller, +Message) is det.

validate_type_(Type, Term, Caller, Message) :-
	validate_(
		error:has_type(Type, Term),
		type_exception_(Type, Term, Caller, Message)
	).

validate_domain_(Type, Constraint, Term, Caller, Message) :-
	validate_type_(Type, Term, Caller, Message),
	validate_(
		call(Constraint, Term),
		domain_exception_(Type, Constraint, Term, Caller, Message)
	).

validate_condition_(Condition, Error, Caller, Message) :-
	validate_(
		call(Condition),
		custom_exception_(Error, Caller, Message)
	).

%%	validate_(:Validate, :CreateEx) is det.

validate_(Validate, _) :-
	call(Validate),
	!.
validate_(_, CreateEx) :-
	call(CreateEx, Exception),
	throw(Exception).

%%	type_exception_(+Type, @Term, :Caller, +Message, -Exception) is det.
%%	domain_exception_(+Type, :Constraint, @Term, :Caller, +Message, -Exception) is det.
%%	custom_exception_(+Error, :Caller, +Message, -Exception) is det.

type_exception_(Type, Term, Caller, Message, Exception) :-
	type_error_(Type, Term, Error),
	custom_exception_(Error, Caller, Message, Exception).

domain_exception_(Type, Constraint, Term, Caller, Message, Exception) :-
	domain_error_(Type, Constraint, Term, Error),
	custom_exception_(Error, Caller, Message, Exception).

custom_exception_(Error, Caller, Message, Exception) :-
	Exception = error(Error, context(Caller, Message)).

%%	type_error_(+Type, @Term, -Error) is det.
%%	domain_error_(+Type, :Constraint, @Term, -Error) is det.

type_error_(Type, Term, Error) :-
	Error = type_error(Type, Term).

domain_error_(Type, Constraint, Term, Error) :-
	Domain = constr(Type, Constraint),
	Error = domain_error(Domain, Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
